"""
Jaime Sendra Berenguer-2020.
MLearner Machine Learning Library Extensions
Author:Jaime Sendra Berenguer<www.linkedin.com/in/jaisenbe>
License: MIT
"""

from .ml_neural import Neural, Neural_sklearn

__all__ = ["Neural", "Neural_sklearn"]
