"""
Jaime Sendra Berenguer-2020.
MLearner Machine Learning Library Extensions
Author:Jaime Sendra Berenguer<www.linkedin.com/in/jaisenbe>
License: MIT
"""

from .load_data_images import load_image, plot_image, plot_image2

__all__ = ["load_image", "plot_image", "plot_image2"]
