
"""Jaime Sendra Berenguer-2020.
MLearner Machine Learning Library Extensions
Author:Jaime Sendra Berenguer <www.linkedin.com/in/jaisenbe>
License: MIT
"""


# from .voting import MajorityVoteClassifier


# __all__ = ["MajorityVoteClassifier"]
